-------------------------------------------------------------------------------
--
-- Simple VGA raster display
-- Written for 128-bit AES decryption project
-- Course: CSEE 4840 - Embedded System Design, Spring 2008
-- Authors: Shrivathsa Bhargav (sb2784)
--			Larry Chen (lc2454)
--			Abhinandan Majumdar (am2993)
--			Shiva Ramudit (syr9)
--
-- Originally authored by Stephen A. Edwards (sedwards@cs.columbia.edu)
--
-- Last modified: 5-8-2008
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity de2_vga_raster is
  
  port (
    clk   : in std_logic;            -- Should be 25.125 MHz

    -- SRAM signals
    address_out : out std_logic_vector(17 downto 0);
    data_in		: in std_logic_vector(15 downto 0);	

    -- Inside rectangle
    rectangle_out : out std_logic;

    VGA_CLK,                         -- Clock
    VGA_HS,                          -- H_SYNC
    VGA_VS,                          -- V_SYNC
    VGA_BLANK,                       -- BLANK
    VGA_SYNC : out std_logic;        -- SYNC
    VGA_R,
    VGA_G,
    VGA_B : out std_logic_vector(9 downto 0) -- R=G=B[9:0]
    );

end de2_vga_raster;

architecture de2_vga_raster_arch of de2_vga_raster is

  -- Video parameters
  
  constant HTOTAL       : integer := 800;
  constant HSYNC        : integer := 96;
  constant HBACK_PORCH  : integer := 48;
  constant HACTIVE      : integer := 640;
  constant HFRONT_PORCH : integer := 16;
  
  constant VTOTAL       : integer := 525;
  constant VSYNC        : integer := 2;
  constant VBACK_PORCH  : integer := 33;
  constant VACTIVE      : integer := 480;
  constant VFRONT_PORCH : integer := 10;

  --Positioning the frame at the top-left of the screen
  constant RECTANGLE_HSTART : integer := 0;
  constant RECTANGLE_HEND   : integer := 319;
  constant RECTANGLE_VSTART : integer := 0;
  constant RECTANGLE_VEND   : integer := 239;

constant header : integer := 539;


  -- Signals for the video controller
  signal Hcount : std_logic_vector(9 downto 0);  -- Horizontal position (0-800)
  signal Vcount : std_logic_vector(9 downto 0);  -- Vertical position (0-524)
  signal EndOfLine, EndOfField : std_logic;

  signal vga_hblank, vga_hsync,vga_vblank, vga_vsync : std_logic;  -- Sync. signals
  signal VGA_I : std_logic_vector(9 downto 0);	-- Grayscale intensity (=R=G=B)

  signal rectangle_h, rectangle_v, rectangle : std_logic;  -- rectangle area

  --RAM signals (to interface with the internally instantiated SRAM)
  signal read						  : std_logic;

  signal reset : std_logic := '0';
 
  signal row : integer := 0;
  signal col : integer := 0;
  signal addr : integer := 0;

  signal row_i, col_i : std_logic_vector(9 downto 0);
  signal data : std_logic_vector(7 downto 0);

begin
   -- Horizontal and vertical counters
  HCounter : process (clk)
  begin
    if rising_edge(clk) then      
      if reset = '1' then
        Hcount <= (others => '0');
      elsif EndOfLine = '1' then
        Hcount <= (others => '0');
      else
        Hcount <= Hcount + 1;
      end if;      
    end if;
  end process HCounter;

  EndOfLine <= '1' when Hcount = HTOTAL - 1 else '0';
  
  VCounter: process (clk)
  begin
    if rising_edge(clk) then      
      if reset = '1' then
        Vcount <= (others => '0');
      elsif EndOfLine = '1' then
        if EndOfField = '1' then
          Vcount <= (others => '0');
        else
          Vcount <= Vcount + 1;
        end if;
      end if;
    end if;
  end process VCounter;

  EndOfField <= '1' when Vcount = VTOTAL - 1 else '0';

  -- State machines to generate HSYNC, VSYNC, HBLANK, and VBLANK
  HSyncGen : process (clk)
  begin
    if rising_edge(clk) then     
      if reset = '1' or EndOfLine = '1' then
        vga_hsync <= '1';
      elsif Hcount = HSYNC - 1 then
        vga_hsync <= '0';
      end if;
    end if;
  end process HSyncGen;
  
  HBlankGen : process (clk)
  begin
    if rising_edge(clk) then
      if reset = '1' then
        vga_hblank <= '1';
      elsif Hcount = HSYNC + HBACK_PORCH then
        vga_hblank <= '0';
      elsif Hcount = HSYNC + HBACK_PORCH + HACTIVE then
        vga_hblank <= '1';
      end if;      
    end if;
  end process HBlankGen;

  VSyncGen : process (clk)
  begin
    if rising_edge(clk) then
      if reset = '1' then
        vga_vsync <= '1';
      elsif EndOfLine ='1' then
        if EndOfField = '1' then
          vga_vsync <= '1';
        elsif Vcount = VSYNC - 1 then
          vga_vsync <= '0';
        end if;
      end if;      
    end if;
  end process VSyncGen;

  VBlankGen : process (clk)
  begin
    if rising_edge(clk) then    
      if reset = '1' then
        vga_vblank <= '1';
      elsif EndOfLine = '1' then
        if Vcount = VSYNC + VBACK_PORCH - 1 then
          vga_vblank <= '0';
        elsif Vcount = VSYNC + VBACK_PORCH + VACTIVE - 1 then
          vga_vblank <= '1';
        end if;
      end if;
    end if;
  end process VBlankGen;

  -- Rectangle generator
  RectangleHGen : process (clk)
  begin
    if rising_edge(clk) then     
      if reset = '1' or Hcount = HSYNC + HBACK_PORCH + RECTANGLE_HSTART - 1 then
        rectangle_h <= '1';
      elsif Hcount = HSYNC + HBACK_PORCH + RECTANGLE_HEND - 1 then
        rectangle_h <= '0';
      end if;      
    end if;
  end process RectangleHGen;

  RectangleVGen : process (clk)
  begin
    if rising_edge(clk) then
      if reset = '1' then       
        rectangle_v <= '0';
      elsif EndOfLine = '1' then
        if Vcount = VSYNC + VBACK_PORCH - 1 + RECTANGLE_VSTART then
          rectangle_v <= '1';
        elsif Vcount = VSYNC + VBACK_PORCH - 1 + RECTANGLE_VEND then
          rectangle_v <= '0';
        end if;
      end if;      
    end if;
  end process RectangleVGen;

  rectangle <= rectangle_h and rectangle_v;

  -- Registered video signals going to the video DAC
  row <= conv_integer(Hcount) - (HSYNC + HBACK_PORCH + RECTANGLE_HSTART);
  row_i <= conv_std_logic_vector(row,10);
  col <= conv_integer(Vcount) - (VSYNC + VBACK_PORCH + RECTANGLE_VSTART);
  col_i <= conv_std_logic_vector(col,10);
  addr <= (col * ((RECTANGLE_HEND - RECTANGLE_HSTART + 1) / 2)) + conv_integer(row_i(9 downto 1)) + header when rectangle = '1' else
           0;

  address_out <= conv_std_logic_vector(addr,18);	

  data <= data_in(15 downto 8) when row_i(0) = '0' else
            data_in(7 downto 0) when row_i(0) = '1' else
            (others => '0');

  VideoOut: process (clk, reset)
  
  begin
    if reset = '1' then
      VGA_R <= "0000000000";
      VGA_G <= "0000000000";
      VGA_B <= "0000000000";
    elsif clk'event and clk = '1' then
      if rectangle = '1' then
        VGA_R <= data & "00";
        VGA_G <= data & "00";
        VGA_B <= data & "00";
--		end if;
      elsif vga_hblank = '0' and vga_vblank ='0' then
        VGA_R <= "0000000000";
        VGA_G <= "0000000000";
        VGA_B <= "0000000000";
      else
        VGA_R <= "0000000000";
        VGA_G <= "0000000000";
        VGA_B <= "0000000000";    
      end if;
    end if;
  end process VideoOut;


  VGA_CLK <= clk;
  VGA_HS <= not vga_hsync;
  VGA_VS <= not vga_vsync;
  VGA_SYNC <= '0';
  VGA_BLANK <= not (vga_hsync or vga_vsync);

  --There's some slack on the right-edge and bottom edge of the screen
  --This is to prevent Nios from attempting a 128-bit block into SRAM at the very right or bottom of the screen
  --  and end up using the SRAM when the VGA wanders back into the image frame
  rectangle_out <= '1' when ((col_i > RECTANGLE_VEND) or ((row_i > RECTANGLE_HEND) and (row_i < (RECTANGLE_HEND + 256)))) else
                   '0';

end de2_vga_raster_arch;
